/*
 *      lark-espeak.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <espeak/speak_lib.h>
#include <string.h>

#include "lark-espeak.h"

#define LARK_ESPEAK_FORMAT      LARK_GSTREAMER_SRC_FORMAT_S16LE
#define LARK_ESPEAK_SAMPLERATE  16000
#define LARK_ESPEAK_CHANNELS    1

struct _LarkEspeak
{
	GObject parent_instance;
	/*Data*/
	gboolean *genstate;
};

struct _LarkEspeakClass
{
	GObjectClass parent_class;
};

G_DEFINE_TYPE (LarkEspeak, lark_espeak, LARK_TYPE_TTS_ENGINE)


static void lark_espeak_voice_profile_change_notify(GObject *gobject, GParamSpec *pspec, LarkEspeak *self)
{
	espeak_VOICE selvoice;
	espeak_ERROR speakErr;
	
	g_return_if_fail(self != NULL);
	
	selvoice.name = lark_tts_engine_get_voice_profile(LARK_TTS_ENGINE(self));
	selvoice.languages = NULL;
	selvoice.gender = 0;
	selvoice.age = 0;
	selvoice.variant = 0;
	
	speakErr = espeak_SetVoiceByProperties(&selvoice);
	
	if (speakErr != EE_OK) {
		g_warning("Error while setting ESpeak voice profile: %i", speakErr);
	}
}

static void lark_espeak_pitch_change_notify(GObject *gobject, GParamSpec *pspec, LarkEspeak *self)
{
	espeak_ERROR speakErr;
	
	g_return_if_fail(self != NULL);
	
	speakErr = espeak_SetParameter(espeakPITCH, (gint)(lark_tts_engine_get_pitch(LARK_TTS_ENGINE(self)) + 1) * 50, FALSE);
	
	if (speakErr != EE_OK) {
		g_warning("Error while setting ESpeak pitch: %i", speakErr);
	}
}

static void lark_espeak_rate_change_notify(GObject *gobject, GParamSpec *pspec, LarkEspeak *self)
{
	espeak_ERROR speakErr;
	
	g_return_if_fail(self != NULL);
	
	speakErr = espeak_SetParameter(espeakRATE, (gint)((lark_tts_engine_get_rate(LARK_TTS_ENGINE(self)) + 1) * 90) + 80, FALSE);
	
	if (speakErr != EE_OK) {
		g_warning("Error while setting ESpeak rate: %i", speakErr);
	}
}

static void lark_espeak_volume_change_notify(GObject *gobject, GParamSpec *pspec, LarkEspeak *self)
{
	espeak_ERROR speakErr;
	
	g_return_if_fail(self != NULL);
	
	speakErr = espeak_SetParameter(espeakVOLUME, (gint)(lark_tts_engine_get_volume(LARK_TTS_ENGINE(self)) + 1) * 100, FALSE);
	
	if (speakErr != EE_OK) {
		g_warning("Error while setting ESpeak volume: %i", speakErr);
	}
}

static void lark_espeak_capitals_indication_change_notify(GObject *gobject, GParamSpec *pspec, LarkEspeak *self)
{
	espeak_ERROR speakErr;
	
	g_return_if_fail(self != NULL);
	
	switch (lark_tts_engine_get_capitals_indication(LARK_TTS_ENGINE(self))) {
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH:
			speakErr = espeak_SetParameter(espeakCAPITALS, 50, FALSE); //50 Hz
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME:
			speakErr = espeak_SetParameter(espeakCAPITALS, 1, FALSE);  //Sound icon
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF:
			speakErr = espeak_SetParameter(espeakCAPITALS, 0, FALSE);  //None
			break;
		default:
			speakErr = espeak_SetParameter(espeakCAPITALS, 1, FALSE);  //Sound icon
			break;
	}
	
	if (speakErr != EE_OK) {
		g_warning("Error while setting ESpeak capitals indication: %i", speakErr);
	}
}

static void lark_espeak_punctuation_indication_change_notify(GObject *gobject, GParamSpec *pspec, LarkEspeak *self)
{
	espeak_ERROR speakErr;
	
	g_return_if_fail(self != NULL);
	
	switch (lark_tts_engine_get_punctuation_indication(LARK_TTS_ENGINE(self))) {
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON:
			speakErr = espeak_SetParameter(espeakPUNCTUATION, espeakPUNCT_ALL, FALSE);
			break;
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF:
			speakErr = espeak_SetParameter(espeakPUNCTUATION, espeakPUNCT_NONE, FALSE);
			break;
		default:
			speakErr = espeak_SetParameter(espeakPUNCTUATION, espeakPUNCT_NONE, FALSE);
			break;
	}
	
	if (speakErr != EE_OK) {
		g_warning("Error while setting ESpeak punctuation indication: %i", speakErr);
	}
}

static void lark_espeak_constructed(GObject *object)
{
	LarkEspeak *self = (LarkEspeak *)object;
	
	g_signal_connect(G_OBJECT(self), "notify::voice-profile", G_CALLBACK(lark_espeak_voice_profile_change_notify), self);
	
	g_signal_connect(G_OBJECT(self), "notify::pitch", G_CALLBACK(lark_espeak_pitch_change_notify), self);
	g_signal_connect(G_OBJECT(self), "notify::rate", G_CALLBACK(lark_espeak_rate_change_notify), self);
	g_signal_connect(G_OBJECT(self), "notify::volume", G_CALLBACK(lark_espeak_volume_change_notify), self);
	
	g_signal_connect(G_OBJECT(self), "notify::capitals-indication", G_CALLBACK(lark_espeak_capitals_indication_change_notify), self);
	g_signal_connect(G_OBJECT(self), "notify::punctuation-indication", G_CALLBACK(lark_espeak_punctuation_indication_change_notify), self);
		
	G_OBJECT_CLASS(lark_espeak_parent_class)->constructed(object);
}

static void lark_espeak_finalize(GObject *object)
{
	LarkEspeak *self __attribute__((unused)) = (LarkEspeak *)object;
	
	G_OBJECT_CLASS(lark_espeak_parent_class)->finalize(object);
}

static void lark_espeak_class_init(LarkEspeakClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->constructed = lark_espeak_constructed;
	object_class->finalize = lark_espeak_finalize;
}

static gint lark_espeak_fill_buffer(gshort *samples, gint count, espeak_EVENT *events)
{
	LarkEspeak *self = (LarkEspeak *)events->user_data;
	
	g_return_val_if_fail(self != NULL, 0);
	
	if (count > 0) {
		g_signal_emit_by_name(G_OBJECT(self), "samples-generated", count, samples);
	}
	
	return (gint)*self->genstate;
}

static void lark_espeak_init(LarkEspeak *self)
{
	espeak_VOICE **voices;
	gint i, l, laststart;
	GSList *voicelist;
	LarkTtsEngineVoice *voice;
	
	/*Create new TTS engine instance*/
	espeak_Initialize(AUDIO_OUTPUT_SYNCHRONOUS, 0, NULL, 0);
	
	/*Callback*/
	espeak_SetSynthCallback(lark_espeak_fill_buffer);
	
	/*Get languages list*/
	voicelist = NULL;
	
	voices = (espeak_VOICE **)espeak_ListVoices(NULL);
	
	for (i=0; voices[i] != NULL; i++) {
		/*New voice entry*/
		voice = g_new0(LarkTtsEngineVoice, 1);
		voice->name = g_strdup(voices[i]->name);
		voice->languages = g_ptr_array_new_with_free_func((GDestroyNotify)g_free);
		voice->genders = g_array_new(FALSE, TRUE, sizeof(gint));
		/*Languages list*/
		laststart = 0;
		for (l = 0; ; l++) {
			if (voices[i]->languages[l] == '\0') {
				g_ptr_array_add(voice->languages, g_strdup(voices[i]->languages + laststart + 1));
				g_array_append_val(voice->genders, voices[i]->gender);
				laststart = l + 1;
				/*List termination*/
				if (voices[i]->languages[l + 1] == '\0') {
					break;				
				}
			}
		}
		/*Add voice entry to list*/
		voicelist = g_slist_prepend(voicelist, voice);
	}
	
	voicelist = g_slist_reverse(voicelist);
	
	g_object_set(self, "voices", voicelist, "format", LARK_ESPEAK_FORMAT, "sample-rate", LARK_ESPEAK_SAMPLERATE, "channels", LARK_ESPEAK_CHANNELS, NULL);
}

gboolean lark_espeak_generate_samples(LarkEspeak *self, gchar *line, gboolean *genstate)
{
	espeak_ERROR speakErr;
	
	g_return_val_if_fail(self != NULL, FALSE);
	g_return_val_if_fail(line != NULL, FALSE);
	g_return_val_if_fail(genstate != NULL, FALSE);
	
	self->genstate = genstate;
	
	speakErr = espeak_Synth(line, strlen(line) + 1, 0, POS_CHARACTER, 0, espeakCHARS_AUTO | espeakPHONEMES | espeakENDPAUSE, NULL, self);
	
	if (speakErr != EE_OK) {
		return FALSE;
	}
	
	return TRUE;
}

LarkEspeak *lark_espeak_new(void)
{
	return g_object_new(lark_espeak_get_type(), NULL);
}
